Index: wp-includes/default-constants.php
===================================================================
--- wp-includes/default-constants.php	(revision 18619)
+++ wp-includes/default-constants.php	(working copy)
@@ -268,6 +268,12 @@
 
 	if ( !defined('WP_POST_REVISIONS') )
 		define('WP_POST_REVISIONS', true);
+
+	/**
+	 * @since 3.3.0
+	 */
+	if ( !defined( 'WP_CRON_LOCK_TIMEOUT' ) )
+		define('WP_CRON_LOCK_TIMEOUT', 60);  // In seconds
 }
 
 /**
Index: wp-includes/cron.php
===================================================================
--- wp-includes/cron.php	(revision 18619)
+++ wp-includes/cron.php	(working copy)
@@ -204,13 +204,13 @@
 	* multiple processes on multiple web servers can run this code concurrently
 	* try to make this as atomic as possible by setting doing_cron switch
 	*/
-	$flag = get_transient('doing_cron');
+	$lock = get_transient('doing_cron');
 
-	if ( $flag > $local_time + 10*60 )
-		$flag = 0;
+	if ( $lock > $local_time + 10*60 )
+		$lock = 0;
 
 	// don't run if another process is currently running it or more than once every 60 sec.
-	if ( $flag + 60 > $local_time )
+	if ( $lock + WP_CRON_LOCK_TIMEOUT > $local_time )
 		return;
 
 	//sanity check
@@ -227,6 +227,7 @@
 			return;
 
 		set_transient( 'doing_cron', $local_time );
+		$lock = $local_time;
 
 		ob_start();
 		wp_redirect( add_query_arg('doing_wp_cron', '', stripslashes($_SERVER['REQUEST_URI'])) );
@@ -241,8 +242,9 @@
 	}
 
 	set_transient( 'doing_cron', $local_time );
+	$lock = $local_time;
 
-	$cron_url = get_option( 'siteurl' ) . '/wp-cron.php?doing_wp_cron';
+	$cron_url = get_option( 'siteurl' ) . '/wp-cron.php?lock=' . $lock;
 	wp_remote_post( $cron_url, array('timeout' => 0.01, 'blocking' => false, 'sslverify' => apply_filters('https_local_ssl_verify', true)) );
 }
 
Index: wp-cron.php
===================================================================
--- wp-cron.php	(revision 18619)
+++ wp-cron.php	(working copy)
@@ -26,6 +26,25 @@
 	require_once('./wp-load.php');
 }
 
+// Uncached doing_cron transient fetch
+function _get_cron_lock() {
+	global $_wp_using_ext_object_cache, $wpdb;
+
+	$value = 0;
+	if ( $_wp_using_ext_object_cache ) {
+		// Clear the internal cache for the doing_cron transient to force a cache refresh
+		if ( function_exists( 'wp_cache_reset' ) )
+			wp_cache_reset( 'transient', 'doing_cron' );
+		$value = wp_cache_get( 'doing_cron', 'transient' );
+	} else {
+		$row = $wpdb->get_row( $wpdb->prepare( "SELECT option_value FROM $wpdb->options WHERE option_name = %s LIMIT 1", '_transient_doing_cron' ) );
+		if ( is_object( $row ) )
+			$value = $row->option_value;
+	}
+
+	return $value;
+}
+
 if ( false === $crons = _get_cron_array() )
 	die();
 
@@ -35,26 +54,52 @@
 if ( isset($keys[0]) && $keys[0] > $local_time )
 	die();
 
-foreach ($crons as $timestamp => $cronhooks) {
+$doing_cron = get_transient( 'doing_cron');
+
+// Use global $lock otherwise GET lock. If no lock, trying grabbing a new lock.
+if ( empty( $lock ) ) {
+	if ( empty( $_GET[ 'lock' ] ) ) {
+		// Called from external script/job. Try setting a lock.
+		if ( $doing_cron && ( $doing_cron + WP_CRON_LOCK_TIMEOUT > $local_time ) )
+			return;
+		$doing_cron = $lock = time();
+		set_transient( 'doing_cron', $lock );
+	} else {
+		$lock = $_GET[ 'lock' ];
+	}
+}
+
+// Check lock
+if ( $doing_cron != $lock )
+	return;
+
+foreach ( $crons as $timestamp => $cronhooks ) {
 	if ( $timestamp > $local_time )
 		break;
 
-	foreach ($cronhooks as $hook => $keys) {
+	foreach ( $cronhooks as $hook => $keys ) {
 
-		foreach ($keys as $k => $v) {
+		foreach ( $keys as $k => $v ) {
 
 			$schedule = $v['schedule'];
 
-			if ($schedule != false) {
+			if ( $schedule != false ) {
 				$new_args = array($timestamp, $schedule, $hook, $v['args']);
 				call_user_func_array('wp_reschedule_event', $new_args);
 			}
 
-			wp_unschedule_event($timestamp, $hook, $v['args']);
+			wp_unschedule_event( $timestamp, $hook, $v['args'] );
 
- 			do_action_ref_array($hook, $v['args']);
+ 			do_action_ref_array( $hook, $v['args'] );
+
+			// If the hook ran too long and another cron process stole the lock, quit.
+			if ( _get_cron_lock() != $lock )
+				return;
 		}
 	}
 }
 
+if ( _get_cron_lock() == $lock )
+	delete_transient( 'doing_cron' );
+
 die();

Index: wp-includes/cache.php
===================================================================
--- wp-includes/cache.php	(revision 18619)
+++ wp-includes/cache.php	(working copy)
@@ -208,10 +208,10 @@
  *
  * @since 2.6.0
  */
-function wp_cache_reset() {
+function wp_cache_reset( $group = '', $key = '' ) {
 	global $wp_object_cache;
 
-	return $wp_object_cache->reset();
+	return $wp_object_cache->reset( $group, $key );
 }
 
 /**
@@ -480,11 +480,23 @@
 	 *
 	 * @since 3.0.0
 	 */
-	function reset() {
-		// Clear out non-global caches since the blog ID has changed.
-		foreach ( array_keys($this->cache) as $group ) {
-			if ( !in_array($group, $this->global_groups) )
-				unset($this->cache[$group]);
+	function reset( $group = '', $id = '' ) {
+		if ( ! empty( $group ) ) {
+			if ( !empty( $id ) ) {
+				// Clear out a particular key
+				if ( isset( $this->cache[$group][$id] ) )
+					unset( $this->cache[$group][$id] );
+			} else {
+				// Clear out a particular group
+				if ( isset( $this->cache[$group] ) )
+					unset( $this->cache[$group] );
+			}
+		} else {
+			// Clear out non-global caches since the blog ID has changed.
+			foreach ( array_keys($this->cache) as $group ) {
+				if ( !in_array($group, $this->global_groups) )
+					unset($this->cache[$group]);
+			}
 		}
 	}
 